import logging
import time
from typing import List

import requests
from aiohttp import ClientSession
from empiric.core.entry import SpotEntry
from empiric.core.utils import currency_pair_to_pair_id
from empiric.publisher.assets import EmpiricAsset
from empiric.publisher.types import PublisherInterfaceT

logger = logging.getLogger(__name__)


class GeminiFetcher(PublisherInterfaceT):
    BASE_URL: str = "https://api.gemini.com/v1"
    SOURCE: str = "GEMINI"

    publisher: str

    def __init__(self, assets: List[EmpiricAsset], publisher):
        self.assets = assets
        self.publisher = publisher

    async def fetch(self, session: ClientSession) -> List[SpotEntry]:
        entries = []
        async with session.get(self.BASE_URL + "/pricefeed") as resp:
            result_json = await resp.json()
            for asset in self.assets:
                if asset["type"] != "SPOT":
                    logger.debug(f"Skipping Gemini for non-spot asset {asset}")
                    continue

                pair = asset["pair"]
                pair_id = currency_pair_to_pair_id(*pair)
                timestamp = int(time.time())
                result = [e for e in result_json if e["pair"] == "".join(pair)]

                if len(result) == 0:
                    logger.debug(f"No entry found for {pair_id} from Gemini")
                    continue

                if len(result) > 1:
                    raise ValueError(
                        f"Found more than one matching entries for Gemini response and price pair {pair}"
                    )

                price = float(result[0]["price"])
                price_int = int(price * (10 ** asset["decimals"]))

                logger.info(f"Fetched price {price} for {pair_id} from Gemini")

                entries.append(
                    SpotEntry(
                        pair_id=pair_id,
                        price=price_int,
                        timestamp=timestamp,
                        source=self.SOURCE,
                        publisher=self.publisher,
                    )
                )
            return entries

    def fetch_sync(self) -> List[SpotEntry]:
        entries = []
        resp = requests.get(self.BASE_URL + "/pricefeed")
        result_json = resp.json()
        for asset in self.assets:
            if asset["type"] != "SPOT":
                logger.debug(f"Skipping Gemini for non-spot asset {asset}")
                continue

            pair = asset["pair"]
            pair_id = currency_pair_to_pair_id(*pair)
            timestamp = int(time.time())
            result = [e for e in result_json if e["pair"] == "".join(pair)]

            if len(result) == 0:
                logger.debug(f"No entry found for {pair_id} from Gemini")
                continue

            if len(result) > 1:
                raise ValueError(
                    f"Found more than one matching entries for Gemini response and price pair {pair}"
                )

            price = float(result[0]["price"])
            price_int = int(price * (10 ** asset["decimals"]))

            logger.info(f"Fetched price {price} for {pair_id} from Gemini")

            entries.append(
                SpotEntry(
                    pair_id=pair_id,
                    price=price_int,
                    timestamp=timestamp,
                    source=self.SOURCE,
                    publisher=self.publisher,
                )
            )
        return entries



import os
import time
from typing import List

from empiric.core.entry import SpotEntry
from web3 import HTTPProvider, Web3

ORACLE_ADDRESS = "0xc09d042ed2f47297d1e8f010aF03d6f094433D65"
ORACLE_ABI = [
    {
        "inputs": [
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "timestamp",
                                "type": "uint256",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "source",
                                "type": "bytes32",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "publisher",
                                "type": "bytes32",
                            },
                        ],
                        "internalType": "struct IOracle.BaseEntry",
                        "name": "base",
                        "type": "tuple",
                    },
                    {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
                    {"internalType": "uint256", "name": "price", "type": "uint256"},
                    {"internalType": "uint256", "name": "volume", "type": "uint256"},
                ],
                "internalType": "struct IOracle.SpotEntry",
                "name": "spotEntry",
                "type": "tuple",
            }
        ],
        "name": "publishSpotEntry",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
    {
        "inputs": [
            {
                "components": [
                    {
                        "components": [
                            {
                                "internalType": "uint256",
                                "name": "timestamp",
                                "type": "uint256",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "source",
                                "type": "bytes32",
                            },
                            {
                                "internalType": "bytes32",
                                "name": "publisher",
                                "type": "bytes32",
                            },
                        ],
                        "internalType": "struct IOracle.BaseEntry",
                        "name": "base",
                        "type": "tuple",
                    },
                    {"internalType": "bytes32", "name": "pairId", "type": "bytes32"},
                    {"internalType": "uint256", "name": "price", "type": "uint256"},
                    {"internalType": "uint256", "name": "volume", "type": "uint256"},
                ],
                "internalType": "struct IOracle.SpotEntry[]",
                "name": "spotEntries",
                "type": "tuple[]",
            }
        ],
        "name": "publishSpotEntries",
        "outputs": [],
        "stateMutability": "nonpayable",
        "type": "function",
    },
]


class EvmHelper:
    def __init__(
        self,
        publisher,
        sender_address,
        private_key,
        provider_uri="https://zksync2-testnet.zksync.dev",
    ):
        self.w3 = Web3(HTTPProvider(endpoint_uri=provider_uri))
        self.chain_id = 280
        self.oracle = self.w3.eth.contract(
            address=ORACLE_ADDRESS,
            abi=ORACLE_ABI,
        )
        self.publisher = publisher
        self.sender = sender_address or os.environ["SENDER_ADDRESS"]
        self.private_key = private_key or os.environ["PRIVATE_KEY"]

    def publish_spot_entry(
        self,
        pair,
        price,
        source,
        volume=0,
        gas_price=int(1e8),
    ):
        nonce = self.w3.eth.getTransactionCount(self.sender)
        txn = self.oracle.functions.publishSpotEntry(
            {
                "base": {
                    "timestamp": int(time.time()),
                    "source": pair,
                    "publisher": self.publisher,
                },
                "pairId": b"ETH/USD",
                "price": price,
                "volume": volume,
            }
        ).buildTransaction(
            {
                "nonce": nonce,
                "gasPrice": gas_price,
                "chainId": self.chain_id,
                "from": self.sender,
            }
        )

        signed_txn = self.w3.eth.account.signTransaction(
            txn, private_key=self.private_key
        )
        self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        return signed_txn.hash.hex()

    def publish_spot_entries(self, spot_entries: List[SpotEntry], gas_price=int(1e8)):
        nonce = self.w3.eth.getTransactionCount(self.sender)
        txn = self.oracle.functions.publishSpotEntries(spot_entries).buildTransaction(
            {
                "nonce": nonce,
                "gasPrice": int(1e8),
                "chainId": 280,
                "from": self.sender,
            }
        )
        signed_txn = self.w3.eth.account.signTransaction(
            txn, private_key=self.private_key
        )
        self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

        return signed_txn.hash.hex()
